// io.cpp
//
#include "costanti.h"
#include "tipo.h"
#include "es1370.h"
	// spiegazioni delle varie costanti/direttive al processore
	// e link alle specifiche della scheda audio/ mixer
	// sono contenute nel file


//#define BOCHS
////////////////////////////////////////////////////////////////////////////////
//    COSTANTI                                                                //
////////////////////////////////////////////////////////////////////////////////


const natl PRIO = 1000;
const natl LIV = LIV_SISTEMA;


////////////////////////////////////////////////////////////////////////////////
//                        CHIAMATE DI SISTEMA USATE                           //
////////////////////////////////////////////////////////////////////////////////

extern "C" natl activate_pe(void f(int), int a, natl prio, natl liv, natb type);
extern "C" void terminate_p();
extern "C" void sem_wait(natl sem);
extern "C" void sem_signal(natl sem);
extern "C" natl sem_ini(int val);
extern "C" natl activate_p(void f(int), int a, natl prio, natl liv);
extern "C" void wfi();	// [9.1]
extern "C" void abort_p();
extern "C" void writevid_n(int off, const unsigned char* vett, int quanti);
extern "C" void attrvid_n(int off, int quanti, unsigned char bg, unsigned char fg, bool blink);
extern "C" void log(log_sev sev, const char* buf, int quanti);
extern "C" void* mem_alloc(natl dim, natl balign = 4);
extern "C" void mem_free(void *p);
extern "C" addr graphic_mode();


// aggiunte rispetto a io.cpp standard:

extern "C" natw pci_read(natw l, natw regn, natl size);
	// troviamo la scheda audio tramite ricerca tra le periferiche PCI
	// e confronto del vendorID e deviceID con quelli della ES1370
	// abbiamo bisogno di leggere nello spazio di configurazione
		
extern "C" void udelay(natl usec);
	// il CODEC viene modificato scrivendo su un apposito registro;
	// i comandi verranno poi trasmessi serialmente al CODEC
	// in caso di CODEC occupato, è opportuno attendere un tempo di
	// 100 microsecondi, molto minore ai 50ms della semplice delay
	// preferiamo, quindi, l'attesa attiva

extern "C" addr trasforma(addr ind_virt);
extern "C" bool resident(addr ind_virt, natl quanti, bool residente);
extern "C" void delay(natl n);


////////////////////////////////////////////////////////////////////////////////
//                      FUNZIONI GENERICHE DI SUPPORTO                        //
////////////////////////////////////////////////////////////////////////////////

// ingresso di un byte da una porta di IO
extern "C" void inputb(ioaddr reg, natb &a);

// uscita di un byte su una porta di IO
extern "C" void outputb(natb a, ioaddr reg);

// ingresso di una word da una porta di IO
extern "C" void inputw(ioaddr reg, natw &a);

// uscita di una word su una porta di IO
extern "C" void outputw(natw a, ioaddr reg);

// ingresso di una long da una porta di IO
extern "C" void inputl(ioaddr reg, natl &a);	

// uscita di una long su una porta di IO
extern "C" void outputl(natl a, ioaddr reg);

extern "C" void flog(log_sev sev, const char* fmt, ...);
void *memset(void *dest, int c, unsigned int n);
void *memcpy(void *dest, const void *src, unsigned int n);

////////////////////////////////////////////////////////////////////////////////
//                    GESTIONE DELLE INTERFACCE SERIALI [9.2]                 //
////////////////////////////////////////////////////////////////////////////////

enum funz { input_n, input_ln, output_n, output_0 };  // [9.2]

struct interfse_reg {	// [9.2]
	ioaddr iRBR, iTHR, iLSR, iIER, iIIR;
};

struct des_se {		// [9.2]
	interfse_reg indreg;
	natl mutex;
	natl sincr;
	natl cont;
	addr punt;
	funz funzione;
	natb stato;
};

const natl S = 2;
extern "C" des_se com[S];	// [9.2]

void input_com(des_se* p_des);	// [9.2]
void output_com(des_se* p_des);	// [9.2]
void estern_com(int i) // [9.2]
{
	natb r;
	des_se *p_des;
	p_des = &com[i];
	for(;;) {
		inputb(p_des->indreg.iIIR, r);
		if ((r&0x06) == 0x04) 
			input_com(p_des);
		else if ((r&0x06) == 0x02)
			output_com(p_des);
		wfi();
	}
}

void startse_in(des_se *p_des, natb vetti[], natl quanti, funz op); // [9.2.1]
extern "C" void c_readse_n(natl serial, natb vetti[], natl quanti, natb& errore) // [9.2.1]
{
	des_se *p_des;

	// (* le primitive non devono mai fidarsi dei parametri
	if (serial >= S) {
		flog(LOG_WARN, "readse_n con serial=%d", serial);
		abort_p();
	}
	// *)

	p_des = &com[serial];
	sem_wait(p_des->mutex);
	startse_in(p_des, vetti, quanti, input_n);
	sem_wait(p_des->sincr);
	errore = p_des->stato;
	sem_signal(p_des->mutex);
}

extern "C" void c_readse_ln(natl serial, natb vetti[], int& quanti, natb& errore)
{
	des_se *p_des;

	// (* le primitive non devono mai fidarsi dei parametri
	if (serial >= S) {
		flog(LOG_WARN, "readse_ln con serial=%d", serial);
		abort_p();
	}
	// *)

	p_des = &com[serial];
	sem_wait(p_des->mutex);
	startse_in(p_des, vetti, 80, input_ln);
	sem_wait(p_des->sincr);
	quanti = p_des->cont;
	errore = p_des->stato;
	sem_signal(p_des->mutex);
}

extern "C" void go_inputse(ioaddr i_ctr);
void startse_in(des_se *p_des, natb vetti[], natl quanti, funz op) // [9.2.1]
{
	p_des->cont = quanti;
	p_des->punt = vetti;
	p_des->funzione = op;
	go_inputse(p_des->indreg.iIER);
}

extern "C" void halt_inputse(ioaddr i_ctr);
void input_com(des_se *p_des) // [9.2.1]
{
	natb c; bool fine;
	fine = false;

	halt_inputse(p_des->indreg.iIER);

	inputb(p_des->indreg.iLSR, c);

	p_des->stato = c & 0x1e;
	if (p_des->stato != 0)
		fine = true;
	else {
		inputb(p_des->indreg.iRBR, c);
		if (p_des->funzione == input_n) {
			*static_cast<natb*>(p_des->punt) = c; // memorizzazione
			p_des->punt = static_cast<natb*>(p_des->punt) + 1;
			p_des->cont--;
			if(p_des->cont == 0)
				fine = true;
		} else {
			if (p_des->funzione == input_ln) {
				if(c == '\r' || c == '\n') {
					fine = true;
					p_des->cont = 80 - p_des->cont;
				} else {
					*static_cast<natb*>(p_des->punt) = c; // memorizzazione
					p_des->punt = static_cast<natb*>(p_des->punt) + 1;
					p_des->cont--;
					if (p_des->cont == 0) {
						fine = true;
						p_des->cont = 80;
					}
				}
			}
		}
	}

	if(fine == true) {
		*static_cast<natb*>(p_des->punt) = 0;	// carattere nullo
		sem_signal(p_des->sincr);
	} else
		go_inputse(p_des->indreg.iIER);
}

void startse_out(des_se *p_des, natb vetto[], natl quanti, funz op);
extern "C" void c_writese_n(natl serial, natb vetto[], natl quanti)	// [9.2.2]
{
	des_se *p_des;

	// (* le primitive non devono mai fidarsi dei parametri
	if (serial >= S) {
		flog(LOG_WARN, "writese_n con serial=%d", serial);
		abort_p();
	}
	// *)

	p_des = &com[serial];
	sem_wait(p_des->mutex);
	startse_out(p_des, vetto, quanti, output_n);
	sem_wait(p_des->sincr);
	sem_signal(p_des->mutex);
}

extern "C" void c_writese_0(natl serial, natb vetto[], natl &quanti)
{
	des_se *p_des;

	// (* le primitive non devono mai fidarsi dei parametri
	if (serial >= S) {
		flog(LOG_WARN, "writese_0 con serial=%d", serial);
		abort_p();
	}
	// *)

	p_des = &com[serial];
	sem_wait(p_des->mutex);
	startse_out(p_des, vetto, 0, output_0);
	sem_wait(p_des->sincr);
	quanti = p_des->cont;
	sem_signal(p_des->mutex);
}

extern "C" void go_outputse(ioaddr i_ctr);
void startse_out(des_se *p_des, natb vetto[], natl quanti, funz op) // [9.2.2]
{
	p_des->cont = quanti;
	p_des->punt = vetto;
	p_des->funzione = op;
	go_outputse(p_des->indreg.iIER);
	output_com(p_des); 
}

extern "C" void halt_outputse(ioaddr i_ctr);
void output_com(des_se *p_des)	// [9.2.2]
{
	natb c; bool fine;
        fine = false;

	if (p_des->funzione == output_n) {
		p_des->cont--;
		if(p_des->cont == 0) {
			fine = true;
			halt_outputse(p_des->indreg.iIER);
		}
		c = *static_cast<natb*>(p_des->punt); //prelievo
		outputb(c, p_des->indreg.iTHR);
		p_des->punt = static_cast<natb*>(p_des->punt) + 1; 
	} else if (p_des->funzione == output_0) {
		c = *static_cast<natb*>(p_des->punt); //prelievo
		if (c == 0) {
			fine = true;
			halt_outputse(p_des->indreg.iIER);
		} else {
			outputb(c, p_des->indreg.iTHR);
			p_des->cont++;
			p_des->punt = static_cast<natb*>(p_des->punt) + 1; 
		}
	}

	if (fine == true)
		sem_signal(p_des->sincr);

}

// ( inizializzazione delle interfacce seriali
extern "C" void com_setup(void);	// vedi "io.S"
// interruzioni hardware delle interfacce seriali
int com_irq[S] = { 4, 3 };

bool com_init()
{
	des_se *p_des;
	natw id;
	natl i, com_base_prio = PRIO;

	com_setup();

	for(i = 0; i < S; ++i) {
		p_des = &com[i];

		if ( (p_des->mutex = sem_ini(1)) == 0xFFFFFFFF) {
			flog(LOG_ERR, "com: impossibile creare mutex");
			return false;
		}
		if ( (p_des->sincr = sem_ini(0)) == 0xFFFFFFFF) {
			flog(LOG_ERR, "com: impossibile creare sincr");
			return false;
		}

		id = activate_pe(estern_com, i, com_base_prio - i, LIV, com_irq[i]);
		if (id == 0xFFFFFFFF) {
			flog(LOG_ERR, "com: impossibile creare proc. esterno");
			return false;
		}

	}
	flog(LOG_INFO, "com: inizializzate %d seriali", S);
	return true;
}
// )


////////////////////////////////////////////////////////////////////////////////
//                         GESTIONE DELLA TASTIERA                            //
////////////////////////////////////////////////////////////////////////////////


struct des_kbd {
	bool escape;
	int pause;
	ioaddr iRBR;
	natl id;
} kbd;

natb pausecode[6] = { 0xe1, 0x1d, 0x45, 0xe1, 0x9d, 0xc5 };

natw kbd_read()
{
	natb c;

	inputb(kbd.iRBR, c);

	if (c == pausecode[kbd.pause]) {
		if (++kbd.pause >= 6) {
			kbd.pause = 0;
			return 0xe17f;
		}
		return 0;
	} else {
		kbd.pause = 0;
	}

	if (c == 0xe0) {
		kbd.escape = true;
		return 0;
	}
	
	if (kbd.escape) {
		kbd.escape = false;
		return 0xe000 | c;
	}

	return c;
}

// smistatore dei caratteri dalla tastiera reale alle tastiere virtuali
void vkbd_putchar(natw code);
void estern_kbd(int h)
{
	natw code;

	for(;;) {
		code = kbd_read();
		if (code) 
			vkbd_putchar(code);				
		wfi();
	}
}
// Interruzione hardware della tastiera
const int KBD_IRQ = 1;

bool kbd_init()
{

	kbd.escape = false;
	kbd.pause = 0;
	kbd.iRBR = (ioaddr)0x60;

	if ( (kbd.id = activate_pe(estern_kbd, 0, PRIO, LIV, KBD_IRQ)) == 0xFFFFFFFF) {
		flog(LOG_ERR, "kbd: impossibile creare estern_kbd");
		return false;
	}

	return true;
}

///////////////////////////////////////////////////////////////////////////////
//    TASTIERE VIRTUALI                                                      //
///////////////////////////////////////////////////////////////////////////////

// sottrazione modulo 'mod' (il '%' del C/C++ calcola il valore sbagliato)
// p1 e p2 devono essere compresi tra 0 e mod-1
inline int circular_sub(int p1, int p2, int mod)
{
	int dist = p1 - p2;
	if (dist < 0) dist += mod;
	return dist;
}

// somma modulo 'mod'
inline int circular_sum(int p1, int p2, int mod)
{
	return (p1 + p2) % mod;
}



const int VKBD_BUF_SIZE = 20;
const int MAX_VKBD = 5;
const int VKBD_CONSOLE = MAX_VKBD - 1;

struct des_vkbd {
	natl mutex;
	natl intr;
	bool intr_enabled;
	bool intr_pending;
	int intr_waiting;
	natw buf[VKBD_BUF_SIZE];
	int first;
	int last;
	int nchar;
	natb leds;
};

des_vkbd array_desvkbd[MAX_VKBD];
des_vkbd *vkbd_active;

bool vkbd_init()
{
	for (int i = 0; i < MAX_VKBD; i++) {
		des_vkbd *k = &array_desvkbd[i];

		if ( (k->mutex = sem_ini(1)) == 0xFFFFFFFF) {
			flog(LOG_ERR, "vkbd: impossibile creare mutex");
			return false;
		}
		if ( (k->intr = sem_ini(0)) == 0xFFFFFFFF) {
			flog(LOG_ERR, "vkbd: impossibile creare intr");
			return false;
		}
		k->intr_enabled = false;
		k->intr_pending = false;
		k->intr_waiting = 0;
		k->first = k->last = k->nchar = 0;
	}
	flog(LOG_INFO, "vkbd: inizializzate %d tastiere virtuali", MAX_VKBD);
	return true;
}

extern "C" void c_vkbd_wfi(int v)
{
	if (v < 0 || v >= MAX_VKBD) {
		flog(LOG_WARN, "(wfi) vkbd inesistente: %d", v);
		abort_p();
	}

	des_vkbd *k = &array_desvkbd[v];

	sem_wait(k->mutex);
	while (!k->intr_pending) {
		k->intr_waiting++;
		sem_signal(k->mutex);
		sem_wait(k->intr);
		sem_wait(k->mutex);
	}
	sem_signal(k->mutex);
}


extern "C" void c_vkbd_switch(int v);
extern "C" void c_vmon_switch(int v);
void vkbd_nuovo_car(des_vkbd* k, natw code, bool clear = false)
{
	sem_wait(k->mutex);

	if (!clear && k->nchar >= VKBD_BUF_SIZE)
		goto out;

	if (clear) 
		k->first = k->last = k->nchar = 0;

	k->buf[k->last] = code;
	k->last = circular_sum(k->last, 1, VKBD_BUF_SIZE);
	k->nchar++;

	if (k->intr_enabled && !k->intr_pending) {
		k->intr_pending = true;
		if (k->intr_waiting > 0) {
			k->intr_waiting--;
			sem_signal(k->intr);
		}
	}
out:
	sem_signal(k->mutex);
}

void vkbd_putchar(natw code)
{
	if (vkbd_active)
		vkbd_nuovo_car(vkbd_active, code);
}

extern "C" void c_vkbd_send(int v, natw code, bool clear)
{
	if (v < -2 || v >= MAX_VKBD) {
		flog(LOG_WARN, "(send) vkbd inesistente: %d", v);
		abort_p();
	}

	des_vkbd *k;

	if (v == -1) {
		for (int i = 0; i < MAX_VKBD; i++) {
			k = &array_desvkbd[i];
			vkbd_nuovo_car(k, code, clear);
		}
	} else {
		if (v == -2) v = VKBD_CONSOLE;
		k = &array_desvkbd[v];
		vkbd_nuovo_car(k, code, clear);
	}
}

extern "C" natw c_vkbd_read(int v)
{
	natw code = 0;

	if (v < 0 || v >= MAX_VKBD) {
		flog(LOG_WARN, "(read) vkbd inesistente: %d", v);
		abort_p();
	}

	des_vkbd *k = &array_desvkbd[v];

	sem_wait(k->mutex);

	if (k->nchar <= 0)
		goto out;

	code = k->buf[k->first];
	k->first = circular_sum(k->first, 1, VKBD_BUF_SIZE);
	k->nchar--;

	k->intr_pending = false;

	if (k->nchar > 0 && k->intr_enabled) {
		k->intr_pending = true;
		if (k->intr_waiting > 0) {
			k->intr_waiting--;
			sem_signal(k->intr);
		}
	}
out:
	sem_signal(k->mutex);
	return code;
}

extern "C" void c_vkbd_intr_enable(int v, bool enable)
{
	if (v < 0 || v >= MAX_VKBD) {
		flog(LOG_WARN, "(intr_en) vkbd inesistente: %d", v);
		abort_p();
	}

	des_vkbd *k = &array_desvkbd[v];

	sem_wait(k->mutex);
	if (enable) {
		if (!k->intr_enabled && k->nchar > 0) {
			k->intr_pending = true;
			if (k->intr_waiting > 0) {
				k->intr_waiting--;
				sem_signal(k->intr);
			}
		}
		k->intr_enabled = true;
	} else {
		k->intr_enabled = false;
	}

	sem_signal(k->mutex);
}

const natl VKBD_LED_CAPSLOCK  = 1L;
const natl VKBD_LED_NUMLOCK   = 2L;
const natl VKBD_LED_SCROLLOCK = 4L;

extern "C" void kbd_set_leds(natb leds);

extern "C" void c_vkbd_leds(int v, natb led, bool on)
{
	if (v < 0 || v >= MAX_VKBD) {
		flog(LOG_WARN, "(leds) vkbd inesistente: %d", v);
		abort_p();
	}

	des_vkbd *k = &array_desvkbd[v];
	if (on) 
		k->leds |= led;
	else
		k->leds &= ~led; 

	if (k == vkbd_active)
		kbd_set_leds(k->leds);
}


extern "C" void c_vkbd_switch(int v)
{
	if (v < 0 || v >= MAX_VKBD) {
		flog(LOG_WARN, "(switch) vkbd inesistente: %d", v);
		abort_p();
	}

	des_vkbd *k = &array_desvkbd[v];

	vkbd_active = k;
	kbd_set_leds(vkbd_active->leds);
}


///////////////////////////////////////////////////////////////////////////////
//  MONITOR VIRTUALI                                                         //
///////////////////////////////////////////////////////////////////////////////

const natl SQRT_NUM_SCREEN = 2;
const natl NUM_SCREEN = 4;

const natl MAX_VMON = NUM_SCREEN;
natw *VIDEO_MEM_BASE = reinterpret_cast<natw *>(0x000b8000);
const int VIDEO_MEM_SIZE = 80 * 25;

const natl FONT_H = 16;
const natl FONT_W = 8;
const natb ch_bitmap[][FONT_H] = {
#include "font.txt"
};
const natl FONT_NUM = sizeof(ch_bitmap) / FONT_H;
natw max_off;

// descrittore di monitor virtuale
struct des_vmon {
	natw* buf;
	natl ch_x, ch_y;
	// text mode
	natw* video;
	int cursor;
	int cursor_size;
	// graph mode
	natl x, y;
	natl w, h;
	natl ubar, dbar;
};

des_vmon array_desvmon[MAX_VMON];
des_vmon *vmon_active;

extern "C" void console_cursor(unsigned int);
extern "C" void console_set_cursor_size(int size);

natb* framebuffer = 0;


static inline void write_pixel(natl x, natl y, natb c) {
	framebuffer[y * VID_W + x] = c;
}

static inline void write_screen(des_vmon* t, natl x, natl y, natb c) {
	write_pixel(t->x + x, t->y + y, c);
}

void write_gc(natl x, natl y, natw v) {
	natb fg = (v >> 8) & 0x0F;
	natb bg = (v >> 12) & 0x07;
	natb c = v;
	if (c >= FONT_NUM) 
		return;
	for (natl i = 0; i < FONT_H; i++) {
		natb w = ch_bitmap[c][i];
		for (natl j = 0; j < FONT_W; j++) {
			write_pixel(x + j, y + i, (w & 0x80) ? fg: bg);
			w <<= 1;
		}
	}
}

void color_bars(des_vmon *t, natb color) {
	//upbar
	for (natl k = 0; k < t->ubar; k++)
		for (natl j = 0; j < t->w; j++) 
			write_screen(t, j, k, color);

	//downbar
	natl downbar = t->h - t->dbar;
	for (natl k = downbar; k < t->h; k++)
		for (natl j = 0; j < t->w; j++) 
			write_screen(t, j, k, color);
}

void framebuffer_cursor(des_vmon *t) {
	natl riga = t->cursor / t->ch_x;
	if (riga > t->ch_y)
		return;
	natl colonna = t->cursor % t->ch_x;
	for (natl i = FONT_H - 1; i >= t->cursor_size; i--) 
		for (natl j = 0; j < FONT_W; j++) 
			framebuffer[(t->y + t->ubar + riga * FONT_H + i) * VID_W +
				    t->x + colonna * FONT_W + j] ^= 0x0F;

}

void framebuffer_update_cursor(des_vmon *t, natl off) {
	if (off == t->cursor)
		return;
	framebuffer_cursor(t);
	t->cursor = off;
	framebuffer_cursor(t);
}


bool vmon_init()
{
	framebuffer = static_cast<natb*>(graphic_mode());
	if (framebuffer)
		flog(LOG_INFO, "modalita' grafica");

	for (natl s = 0; s < MAX_VMON; s++) {
		des_vmon *t = &array_desvmon[s];
		t->cursor = 0;
		t->cursor_size = 14;
		
		if (framebuffer) {
			t->w = VID_W / SQRT_NUM_SCREEN;
			t->h = VID_H / SQRT_NUM_SCREEN;
	 		t->x = t->w * (s % SQRT_NUM_SCREEN);
			t->y = t->h * (s / SQRT_NUM_SCREEN);
			t->ubar = 32;
			t->dbar = 16;
			t->ch_x = t->w / FONT_W;
			t->ch_y = (t->h - t->ubar - t->dbar) / FONT_H;

			color_bars(t, 0x09);	
			////divisori verticali
			//for (int k = 0; k < t->h; k++) {
			//	write_screen(s, 0, k, 0x0B);
			//	write_screen(s, t->w - 1, k, 0x0B);
			//}

			////divisori orizzontali
			//for (int k = 0; k < t->w; k++) {
			//	write_screen(s, k, t->h - 1, 0x0B);
			//	write_screen(s, k, 0, 0x0B);
			//}
			framebuffer_cursor(t);
		} else {
			t->ch_x = 80;
			t->ch_y = 25;
		}
		t->video = t->buf = static_cast<natw*>(mem_alloc((t->ch_x * t->ch_y) * sizeof(natw)));
	} 
	flog(LOG_INFO, "vmon: inizializzati %d monitor virtuali", MAX_VMON);
	return true;
}

static void vmon_put(int v, int off, natw code)
{
	des_vmon *t = &array_desvmon[v];
	if (framebuffer) {
		int riga = off / t->ch_x;
		int colonna = off % t->ch_x;
		write_gc(t->x + colonna * FONT_W,
		         t->y + riga * FONT_H + t->ubar, code);
		if (off == t->cursor)
			framebuffer_cursor(t);
	}
	t->video[off] = code;
}

static inline natw vmon_get(int v, int off)
{
	return array_desvmon[v].video[off];
}

extern "C" void c_vmon_write_n(int v, int off, natw vetti[], int quanti)
{
	if (v < 0 || v >= MAX_VMON) {
		flog(LOG_WARN, "(write) vmon inesistente: %d", v);
		abort_p();
	}

	des_vmon* t = &array_desvmon[v];
	int max_off = t->ch_x * t->ch_y;
	if (quanti <= 0 || off >= max_off || off + quanti <= 0)
		return;

	if (off < 0) {
		vetti += -off;
		quanti -= -off;
		off = 0;
	}
	if (off + quanti >= max_off) 
		quanti = max_off - off;
	for (int i = 0; i < quanti; i++)
		vmon_put(v, off + i, vetti[i]);
}
void light_to(int v)
{
	des_vmon* t = &array_desvmon[v];

	if (vmon_active) 
		color_bars(vmon_active, 0x09);
	color_bars(t, 0x0B);

}

extern "C" void c_vmon_switch(int v)
{
	if (v < 0 || v >= MAX_VMON) {
		flog(LOG_WARN, "(switch) vmon inesistente: %d", v);
		abort_p();
	}
	if (framebuffer) {
		light_to(v);
		vmon_active = &array_desvmon[v];
	} else {
		if (vmon_active) {
			memcpy(vmon_active->buf, VIDEO_MEM_BASE, VIDEO_MEM_SIZE * 2);
			vmon_active->video = vmon_active->buf;
		}
		vmon_active = &array_desvmon[v];
		vmon_active->video = VIDEO_MEM_BASE;
		memcpy(VIDEO_MEM_BASE, vmon_active->buf, VIDEO_MEM_SIZE * 2);
		console_set_cursor_size(vmon_active->cursor_size);
		console_cursor(vmon_active->cursor);
	}
}

extern "C" bool c_vmon_getsize(int v, natw& maxx, natw& maxy)
{
	if (v < 0 || v >= MAX_VMON) {
		return false;
	}
	des_vmon *t = &array_desvmon[v];
	maxx = t->ch_x;
	maxy = t->ch_y;
	return true;
}

			
	
extern "C" void c_vmon_setcursor(int v, natl off)
{
	if (v < 0 || v >= MAX_VMON) {
		flog(LOG_WARN, "(cursor) vmon inesistente: %d", v);
		abort_p();
	}
	des_vmon *t = &array_desvmon[v];
	if (t == vmon_active) {
		if (framebuffer) {
			framebuffer_update_cursor(t, off);
		} else {
			t->cursor = off;
			console_cursor(t->cursor);
		}
	}
}


extern "C" void c_vmon_cursor_shape(int v, int shape)
{
	if (v < 0 || v >= MAX_VMON) {
		flog(LOG_WARN, "(cshape) vmon inesistente: %d", v);
		abort_p();
	}

	int size = 14;
	if (shape == 1) 
		size = 2;
	des_vmon *t = &array_desvmon[v];
	if (framebuffer)
		framebuffer_cursor(t);
	t->cursor_size = size;
	if (framebuffer)
		framebuffer_cursor(t);
	else
		console_set_cursor_size(t->cursor_size);

}

////////////////////////////////////////////////////////////////////////////////
//                         GESTIONE DELLA CONSOLE [9.5]                       //
////////////////////////////////////////////////////////////////////////////////

natw COLS = 80; 	// [9.5]
natw ROWS = 25;	// [9.5]
natl VIDEO_SIZE = COLS * ROWS;	// [9.5]

struct des_vid {
	int vmon;
	natl x, y;
	natb attr;
};

const natl MAX_CODE = 29; // [9.5]

struct des_ckbd { // [9.5]
	int vkbd;
	addr punt;
	natl cont;
	bool shift;
	natb tab[MAX_CODE];
	natb tabmin[MAX_CODE];
	natb tabmai[MAX_CODE];
};

struct des_console { // [9.5]
	natl mutex;
	natl sincr;
	des_ckbd kbd;
	des_vid vid;
};

extern "C" des_console console; // [9.5]

void scroll(des_vid *p_des)	// [9.5]
{
	int vmon = p_des->vmon;
	for (int i = 0; i < VIDEO_SIZE - COLS; i++) 
		vmon_put(vmon, i, vmon_get(vmon, i + COLS));
	for (natl i = 0; i < COLS; i++)
		vmon_put(vmon, VIDEO_SIZE - COLS + i, 0 | p_des->attr << 8);
	p_des->y--;
}

void writeelem(natb c) {	// [9.5]
	des_vid* p_des = &console.vid;
	int vmon = p_des->vmon;
	switch (c) {
	case '\r':
		p_des->x = 0;
		break;
	case '\n':
		p_des->y++;
		if (p_des->y >= ROWS)
			scroll(p_des);
		break;
	case '\b':
		if (p_des->x > 0 || p_des->y > 0) {
			if (p_des->x == 0) {
				p_des->x = COLS - 1;
				p_des->y--;
			} else
				p_des->x--;
		}
		break;
	default:
		vmon_put(vmon, p_des->y * COLS + p_des->x, c | p_des->attr << 8);
		p_des->x++;
		if (p_des->x >= COLS) {
			p_des->x = 0;
			p_des->y++;
		}
		if (p_des->y >= ROWS) 
			scroll(p_des);
		break;
	}
	c_vmon_setcursor(vmon, p_des->y * COLS + p_des->x);
}

void writeseq(cstr seq)	// [9.5]
{
	const natb* pn = static_cast<const natb*>(seq);
	while (*pn != 0) {
		writeelem(*pn);
		pn++;
	}
}

extern "C" void c_writeconsole(cstr buff) // [9.5]
{
	des_console *p_des = &console;
	sem_wait(p_des->mutex);
	writeseq(buff);
	writeseq("\r\n");
	sem_signal(p_des->mutex);
}

//extern "C" void go_inputkbd(interfkbd_reg indreg); // [9.5]
//extern "C" void halt_inputkbd(interfkbd_reg indreg); // [9.5]

void startkbd_in(des_ckbd* p_des, str buff) // [9.5]
{
	p_des->punt = buff;
	p_des->cont = 80;
	//go_inputkbd(p_des->indreg);
	c_vkbd_intr_enable(p_des->vkbd, true);
}

extern "C" void c_readconsole(str buff, natl& quanti) // [9.5]
{
	des_console *p_des;

	p_des = &console;
	sem_wait(p_des->mutex);
	startkbd_in(&p_des->kbd, buff);
	sem_wait(p_des->sincr);
	quanti = p_des->kbd.cont;
	sem_signal(p_des->mutex);
}

natb converti(des_ckbd* p_des, natb c) { // [9.5]
	natb cc;
	natl pos = 0;
	while (pos < MAX_CODE && p_des->tab[pos] != c)
		pos++;
	if (pos == MAX_CODE)
		return 0;
	if (p_des->shift)
		cc = p_des->tabmai[pos];
	else
		cc = p_des->tabmin[pos];
	return cc;
}

void estern_ckbd(int h) // [9.5]
{
	des_console *p_des = &console;
	natw C;
	natb a, c;
	bool fine;

	for(;;) {
		//halt_inputkbd(p_des->kbd.indreg);
		c_vkbd_intr_enable(p_des->kbd.vkbd, false);

		//inputb(p_des->kbd.indreg.iRBR, c);
		C = c_vkbd_read(p_des->kbd.vkbd);
		if (C == 0xFFFF)
			terminate_p();

		c = (natb)C;
		
		fine = false;
		switch (c) {
		case 0x2a: // left shift make code
			p_des->kbd.shift = true;
			break;
		case 0xaa: // left shift break code
			p_des->kbd.shift = false;
			break;
		default:
			if (c < 0x80) {
				a = converti(&p_des->kbd, c);
				if (a == 0)
					break;
				if (a == '\b') {
					if (p_des->kbd.cont < 80) {
						p_des->kbd.punt = static_cast<natb*>(p_des->kbd.punt) - 1;
						p_des->kbd.cont++;
						writeseq("\b \b");
					}
				} else if (a == '\r' || a == '\n') {
					fine = true;
					p_des->kbd.cont = 80 - p_des->kbd.cont;
					*static_cast<natb*>(p_des->kbd.punt) = 0;
					writeseq("\r\n");
				} else {
					*static_cast<natb*>(p_des->kbd.punt) = a;
					p_des->kbd.punt = static_cast<natb*>(p_des->kbd.punt) + 1;
					p_des->kbd.cont--;
					writeelem(a);
					if (p_des->kbd.cont == 0) {
						fine = true;
						p_des->kbd.cont = 80;
					}
				}
			}
			break;
		}
		if (fine == true) 
			sem_signal(p_des->sincr);
		else
			//go_inputkbd(p_des->kbd.indreg);
			c_vkbd_intr_enable(p_des->kbd.vkbd, true);
		//nwfi(master);
		c_vkbd_wfi(p_des->kbd.vkbd);
	}
}

// (* inizializzazioni
bool con_vid_init(int vmon);

extern "C" void c_iniconsole(natb cc)
{
	des_vid *p_des = &console.vid;
	p_des->attr = cc;
	int vmon = p_des->vmon;
	for (int i = 0; i < VIDEO_SIZE; i++) 
		vmon_put(vmon, i, i | p_des->attr << 8);
	p_des->x = p_des->y = 0;
	c_vmon_setcursor(vmon, p_des->y * COLS + p_des->x);
}

bool con_kbd_init(int vkbd)
{
	console.kbd.vkbd = vkbd;
	if (activate_p(estern_ckbd, 0, PRIO, LIV) == 0xFFFFFFFF) {
		flog(LOG_ERR, "console: impossibile creare estern_ckbd");
		return false;
	}
	return true;
}

extern "C" des_vid vid;

bool con_vid_init(int vmon)
{
	c_vmon_getsize(vmon, COLS, ROWS);
	VIDEO_SIZE = COLS * ROWS;
	des_vid *p_des = &console.vid;
	p_des->vmon = vmon;
	for (int i = 0; i < VIDEO_SIZE; i++) 
		vmon_put(vmon, i, 0 | p_des->attr << 8);
	c_vmon_setcursor(vmon, p_des->y * COLS + p_des->x);
	flog(LOG_INFO, "vid: video inizializzato");
	return true;
}

bool console_init(int vkbd, int vmon) {
	des_console *p_des = &console;

	if ( (p_des->mutex = sem_ini(1)) == 0xFFFFFFFF) {
		flog(LOG_ERR, "console: impossibile creare mutex");
		return false;
	}
	if ( (p_des->sincr = sem_ini(0)) == 0xFFFFFFFF) {
		flog(LOG_ERR, "console: impossibile creare sincr");
		return false;
	}
	return con_kbd_init(vkbd) && con_vid_init(vmon);
}

// *)

/////////////////////////////////////////////////////////////////////////////////
//                  FUNZIONI DI LIBRERIA                                       //
/////////////////////////////////////////////////////////////////////////////////
typedef char *va_list;

// Versione semplificata delle macro per manipolare le liste di parametri
//  di lunghezza variabile; funziona solo se gli argomenti sono di
//  dimensione multipla di 4, ma e' sufficiente per le esigenze di printk.
//
#define va_start(ap, last_req) (ap = (char *)&(last_req) + sizeof(last_req))
#define va_arg(ap, type) ((ap) += sizeof(type), *(type *)((ap) - sizeof(type)))
#define va_end(ap)

natl strlen(const char *s)
{
	natl l = 0;

	while(*s++)
		++l;

	return l;
}

char *strncpy(char *dest, const char *src, unsigned long l)
{
	unsigned long i;

	for(i = 0; i < l && src[i]; ++i)
		dest[i] = src[i];

	return dest;
}

static const char hex_map[16] = { '0', '1', '2', '3', '4', '5', '6', '7',
	'8', '9', 'a', 'b', 'c', 'd', 'e', 'f' };

static void htostr(char *buf, unsigned long l)
{
	int i;

	buf[0] = '0';
	buf[1] = 'x';

	for(i = 9; i > 1; --i) {
		buf[i] = hex_map[l % 16];
		l /= 16;
	}
}

static void itostr(char *buf, unsigned int len, long l)
{
	natl i, div = 1000000000, v, w = 0;

	if(l == (-2147483647 - 1)) {
		strncpy(buf, "-2147483648", 12);
		return;
	} else if(l < 0) {
		buf[0] = '-';
		l = -l;
		i = 1;
	} else if(l == 0) {
		buf[0] = '0';
		buf[1] = 0;
		return;
	} else
		i = 0;

	while(i < len - 1 && div != 0) {
		if((v = l / div) || w) {
			buf[i++] = '0' + (char)v;
			w = 1;
		}

		l %= div;
		div /= 10;
	}

	buf[i] = 0;
}

#define DEC_BUFSIZE 12

int vsnprintf(char *str, natl size, const char *fmt, va_list ap)
{
	natl in = 0, out = 0, tmp;
	char *aux, buf[DEC_BUFSIZE];

	while(out < size - 1 && fmt[in]) {
		switch(fmt[in]) {
			case '%':
				switch(fmt[++in]) {
					case 'd':
						tmp = va_arg(ap, int);
						itostr(buf, DEC_BUFSIZE, tmp);
						if(strlen(buf) >
								size - out - 1)
							goto end;
						for(aux = buf; *aux; ++aux)
							str[out++] = *aux;
						break;
					case 'x':
						tmp = va_arg(ap, int);
						if(out > size - 11)
							goto end;
						htostr(&str[out], tmp);
						out += 10;
						break;
					case 's':
						aux = va_arg(ap, char *);
						while(out < size - 1 && *aux)
							str[out++] = *aux++;
						break;	
				}
				++in;
				break;
			default:
				str[out++] = fmt[in++];
		}
	}
end:
	str[out++] = 0;

	return out;
}

int snprintf(char *buf, unsigned long n, const char *fmt, ...)
{
	va_list ap;
	int l;

	va_start(ap, fmt);
	l = vsnprintf(buf, n, fmt, ap);
	va_end(ap);

	return l;
}

// copia n byte da src a dest
void *memcpy(void *dest, const void *src, unsigned int n)
{
	char       *dest_ptr = static_cast<char*>(dest);
	const char *src_ptr  = static_cast<const char*>(src);

	if (src_ptr < dest_ptr && src_ptr + n > dest_ptr)
		for (int i = n - 1; i >= 0; i--)
			dest_ptr[i] = src_ptr[i];
	else
		for (natl i = 0; i < n; i++)
			dest_ptr[i] = src_ptr[i];

	return dest;
}

// scrive n byte pari a c, a partire da dest
void *memset(void *dest, int c, unsigned int n)
{
	char *dest_ptr = static_cast<char*>(dest);

        for (natl i = 0; i < n; i++)
              dest_ptr[i] = static_cast<char>(c);

        return dest;
}

// log formattato
void flog(log_sev sev, const char *fmt, ...)
{
	va_list ap;
	char buf[LOG_MSG_SIZE];

	va_start(ap, fmt);
	int l = vsnprintf(buf, LOG_MSG_SIZE, fmt, ap);
	va_end(ap);

	if (l > 1)
		log(sev, buf, l - 1);
}

////////////////////////////////////////////////////////////////////
//                 SCHEDA AUDIO ES1370	(ENSONIQ PCI)             //
////////////////////////////////////////////////////////////////////

//////////* Sezione dati */////////////////

// struttura per i registri della es1370
struct ens_reg{
	ioaddr control;			// Registro di controllo Interrupt/Chip 
	ioaddr status; 			// 0x04	Registro di stato per Interrupt/Chip 
							// 		sola lettura

	ioaddr uart_data; 		// 0x08	Registro UART
	union{
		ioaddr uart_status; 	
							// 0x09	Registro di stato UART
							//		sola lettura
		ioaddr uart_control; 	
							// 0x09	Registro di controllo UART
							//		sola scrittura
	};
	ioaddr uart_res;		// 0x0a	Registro riservato UART

	ioaddr mem_page;		// 0x0c	Registro per le pagine di memoria
	
	ioaddr codec_1370; 		// 0x10	Registro di scrittura nel CODEC
							//		sola scrittura
							
	ioaddr channel_status; 	// 0x1c   
	ioaddr serial;			// 0x20	Registro di controllo 
							//		per l'interfaccia seriale 

	ioaddr dac1_count; 		// 0x24	Registro di conteggio 
							//		dei campioni per DAC1
	ioaddr dac2_count; 		// 0x28	Registro di conteggio 
							//		dei campioni per DAC2
	ioaddr adc_count;  		// 0x2c	Registro di conteggio 
							//		dei campioni per ADC

	union{
	ioaddr dac1_frame; 		// 0x30	Pagina 0x0c; DAC1 indirizzo buffer
	ioaddr adc_frame;  		// 0x30	Pagina 0x0d; ADC indirizzo buffer
	ioaddr uart_fifo;  		// 0x30	Pagina 0x0e; registro coda UART
	};

	union{
	ioaddr dac1_size;  		// 0x34	Pagina 0x0c; DAC1 dimensione buffer
	ioaddr adc_size;	 	// 0x34	Pagina 0x0d; ADC dimensione buffer
	};

	union{
	ioaddr dac2_frame; 		// 0x38 Pagina 0x0c; DAC2 indirizzo buffer
	ioaddr phantom_frame; 	// 0x38 Pagina 0x0d: Indirizzo buffer "phantom"
	};

	union{
	ioaddr dac2_size;  		// 0x3c	Pagina 0x0c DAC2 dimensione buffer
	ioaddr phantom_count; 	// 0x3c Pagina 0x0d: 
							//		contatore buffer "phantom"
	};
};


// struttura dati per un suono da eseguire
// (descrittore di canale)

struct suono {
	int quanti;		// dimensione (in byte) del file sonoro
	
	natw oldsize; 	// vecchio contatore di longword
	addr puntatore_brano;	// indirizzo virtuale del suono da eseguire
	addr buffer;		// indirizzo virtuale del primo byte del
						// buffer (uguale)
	natl mutex;		// semaforo di mutua esclusione
	natl sincr;		// semaforo di sincronizzazione
};	


// struttura descrittore di scheda audio
struct des_snd {
	ens_reg indreg;
			// indirizzi dei registri
	natl chip_mutex;
			// semaforo di mutua esclusione per il CODEC
	natl es_mutex;
			// semaforo di mutua esclusione per la scheda audio
	natb irq;
			// piedino di interruzione
	bool presente;	
			// è opportuno settare un booleano nel descrittore
			// in questo modo evito che il modulo utente chiami
			// funzioni audio inesistenti			
	natb chip_reg[0x1A];
			// poichè il chip è accessibile in sola scrittura
			// qualsiasi modifica rispetto allo stato di reset
			// deve essere mantenuta in una struttura dati
			// per essere letta, e deve essere protetta
			// tramite semaforo di mutua esclusione

		
	natl ctrl;	// registro di controllo
	natl sctrl;	// registro -seriale- di controllo
	natl cssr;	// registro di stato
	natb uartc;	// registro di controllo UART
	
	// seppur questi registri (al contrario dei registri del CODEC)
	// sono anche leggibili, conviene tenerne una copia in memoria
	// perchè alcuni controlli possano essere fatti anche senza dover
	// accedere alla periferica

	suono canali[3];
			// canali[0] il canale ADC
			// canali[1] il canale DAC1
			// canali[2] il canale DAC2			
};


des_snd ensoniq;


//////////* Sezione testo */////////////////

// la pci_read prende come parametro un valore w (parola)
// formato dalla periferica e dalla funzione nel bus
// di cui vogliamo leggere lo spazio di configurazione

natw make_w(natb dev, natb fun)
{
	natw res= (dev << 3) | fun;
	return res;
}



// alcune funzioni di utilità per scrivere nel buffer circolare
// utilizzato dalla scheda audio
void copia_snd_buffer(addr puntatore_brano, addr buffersuono, 
										int oldsize, int n_long)
{
	int giro=BUFFERSIZE>>2;
	
	int* brano= (int*) (puntatore_brano);
	int* buffer=(int*) (buffersuono);
	for(int i=0; i<n_long; i++)
		buffer[(i+oldsize)%giro]=brano[i];
}

void copia_snd_buffer_B( addr puntatore_brano, addr buffersuono, 
											int oldsize,  int n)
{
	int giro=BUFFERSIZE;
	
	natb* brano = (natb*) (puntatore_brano);
	natb* buffer=(natb*) (buffersuono);
	for(int i=0; i<n; i++)
		buffer[(i+oldsize)%giro]=brano[i];
}


// crea silenzio
void mute_snd_buffer(addr buffersuono, int oldsize, int n_long)
{	
	int giro=BUFFERSIZE>>2;
	int* buffer=(int*) (buffersuono);
	for(int i=0; i<n_long; i++)
		buffer[(i+oldsize)%giro]=0x00;
}
void mute_snd_buffer_B(addr buffersuono, int oldsize, int n)
{
	int giro=BUFFERSIZE;
	
	natb* buffer=(natb*) (buffersuono);
	for(int i=0; i<n; i++)
		buffer[(i+oldsize)%giro]=0x00;
}


#if 0
// FUNZIONI MIXER

// accesso al chip ak4531 per la gestione del mixer

void ak4531_write(natb reg, natb data)
{
	des_snd * p_des;
	p_des = & ensoniq;
	
	natl statuslong;
	natl info = ES_CODEC_WRITE(reg,data);
	
	inputl(p_des->indreg.status, statuslong);

	if((statuslong & ES_1370_CSTAT)==0)
	{
		outputl(info, p_des->indreg.codec_1370);		
		return;
	}
	udelay(100);
	
	inputl(p_des->indreg.status, statuslong);
	if(!(statuslong & (ES_1370_CSTAT))){
		outputl(info , p_des->indreg.codec_1370);
		inputl(p_des->indreg.codec_1370, info);
		return;
	}
	flog(LOG_WARN, "Errore nella scrittura del CODEC; stato attuale: %x", statuslong);
}



// reset del CODEC
void ak4531_reset()
{	
	ak4531_write(AK4531_RESET, 0x03);	
	ak4531_write(AK4531_RESET, 0x02);
}

// inizializzazione del codec
bool ak4531_init(ioaddr base)
{
	des_snd * p_des;
	p_des= &ensoniq;
		
	ak4531_reset();
	
	p_des->chip_reg[AK4531_LMASTER]		= 0x80 ;
	p_des->chip_reg[AK4531_RMASTER]		= 0x80 ;
	p_des->chip_reg[AK4531_LVOICE]		= 0x86 ;
	p_des->chip_reg[AK4531_RVOICE]		= 0x86 ;
	p_des->chip_reg[AK4531_LFM]		= 0x86 ;
	p_des->chip_reg[AK4531_RFM]		= 0x86 ;
	p_des->chip_reg[AK4531_LCD]		= 0x86 ;
	p_des->chip_reg[AK4531_RCD]		= 0x86 ;
	p_des->chip_reg[AK4531_LLINE]		= 0x86 ;
	p_des->chip_reg[AK4531_RLINE]		= 0x86 ;
	p_des->chip_reg[AK4531_LAUXA]		= 0x86 ;
	p_des->chip_reg[AK4531_RAUXA]		= 0x86 ;
	p_des->chip_reg[AK4531_MONO1]		= 0x86 ;
	p_des->chip_reg[AK4531_MONO2]		= 0x86 ;
	p_des->chip_reg[AK4531_MIC]		= 0x86 ;
	p_des->chip_reg[AK4531_MONO_OUT]	= 0x80 ;
	p_des->chip_reg[AK4531_OUT_SW1]		= 0x00 ;
	p_des->chip_reg[AK4531_OUT_SW2]		= 0x00 ;
	p_des->chip_reg[AK4531_LIN_SW1]		= 0x00 ;
	p_des->chip_reg[AK4531_RIN_SW1]		= 0x00 ;
	p_des->chip_reg[AK4531_LIN_SW2]		= 0x00 ;
	p_des->chip_reg[AK4531_RIN_SW2]		= 0x00 ;
	p_des->chip_reg[AK4531_RESET]		= 0x00 ;
	p_des->chip_reg[AK4531_CLOCK]		= 0x03 ;
	p_des->chip_reg[AK4531_AD_IN]		= 0x00 ;
	p_des->chip_reg[AK4531_MIC_GAIN]	= 0x00 ;

//** DA PENSARCI **********************************************	
//	for(int i=AK4531_LMASTER; i<=AK4531_MIC_GAIN; i++)
//		if(i!=AK4531_CLOCK && i!= AK4531_RESET)
//			ak4531_write(i, p_des->chip_reg[i]);
//**************************************************************
				// i dati iniziali sono presi dal foglio di specifiche
				// del CODEC AK4531
				// e equivalgono ai valori al reset asincrono
				// questa parte, perciò, non è strettamente necessaria
				// ma potrebbe essere una scelta oculata
				// per essere certi della coerenza dei valori nel CODEC
				
	if ((p_des->chip_mutex = sem_ini(1)) == 0xFFFFFFFF) {
		flog(LOG_ERR, "com: impossibile creare mutex");
		return false;
		}
	return true;
}


// primitive per il modulo utente
extern "C" void ak4531_setMixer(natb &err, natb port, natb value)
{
	des_snd * p_des;
	p_des= &ensoniq;
	
	if(!p_des->presente)
	{
		flog(LOG_WARN, "Scheda audio non presente!");
		err = -2; 	
		return;
	}
	err=0;
	
	// *(le primitive non devono mai fidarsi dei parametri
	
	if(		port<AK4531_LMASTER || port >AK4531_MIC_GAIN 
		|| 	port==AK4531_RESET 	|| port==AK4531_CLOCK)
	{						
		flog(LOG_WARN, "Tentativo di accedere a una porta non corretta");
		err=-1;
		return;
	}
	
	// e non permettiamo di effettuare il reset o modificare il clock da utente *)

	sem_wait(p_des->es_mutex);	
	sem_wait(p_des->chip_mutex);
	p_des->chip_reg[port]=value;

	ak4531_write(port, value);
	sem_signal(p_des->chip_mutex);
	sem_signal(p_des->es_mutex);
}

extern "C" void ak4531_getMixer(natb &err, natb &value, natb port)
{
	des_snd * p_des;
	p_des= &ensoniq;
	if(!p_des->presente)
	{
		flog(LOG_WARN, "Scheda audio non presente!");
		err = -2; 	
		return;
	}		
	err=0;
	
	// *(le primitive non devono mai fidarsi dei parametri
	
	if(port<AK4531_LMASTER || port >AK4531_MIC_GAIN){
		flog(LOG_WARN, "Tentativo di accedere a una porta non corretta");
		err=-1;
		return;
	}
	//	*)	
	sem_wait(p_des->chip_mutex);
		// poichè leggiamo da memoria, e non dal CODEC (non possiamo)
		// non ha senso bloccare la periferica: facciamo solo sì che
		// durante la lettura nessuno possa scrivere nel registro CODEC

	
	value= p_des->chip_reg[port];
	
	sem_signal(p_des->chip_mutex);
}
#endif

// FUNZIONI INIZIALIZZAZIONE
void gestisci_int_dac(int channel);
void estern_snd(int h);
	// dichiarazione processo esterno

bool ensoniq_init(natb dev, natb fun)
{
	des_snd * p_des;
	p_des = & ensoniq;
	p_des->presente= false;

	natw w= make_w(dev, fun);
	natl base_address = pci_read(w, 0x10, 4);
	base_address = base_address & 0xFFFE; 			
			// resettiamo il bit meno significativo, a uno da specifiche

	p_des->indreg.control		= base_address | ES_CONTROL;		
	p_des->indreg.status 		= base_address | ES_STATUS; 
	p_des->indreg.uart_data 	= base_address | ES_UART_DATA; 
	p_des->indreg.uart_status	= base_address | ES_UART_STATUS; 
	p_des->indreg.uart_res		= base_address | ES_UART_RES;		
	p_des->indreg.mem_page		= base_address | ES_MEM_PAGE;
	p_des->indreg.codec_1370 	= base_address | ES_CODEC_1370;
	p_des->indreg.channel_status= base_address | ES_CHANNEL_STATUS; 
	p_des->indreg.serial		= base_address | ES_SERIAL;	
	p_des->indreg.dac1_count	= base_address | ES_DAC1_COUNT;
	p_des->indreg.dac2_count	= base_address | ES_DAC2_COUNT; 
	p_des->indreg.adc_count		= base_address | ES_ADC_COUNT; 
	p_des->indreg.dac1_frame	= base_address | ES_DAC1_FRAME;
	p_des->indreg.dac1_size		= base_address | ES_DAC1_SIZE;
	p_des->indreg.dac2_frame	= base_address | ES_DAC2_FRAME;
	p_des->indreg.dac2_size		= base_address | ES_DAC2_SIZE;

	
	if ((p_des->es_mutex = sem_ini(1)) == 0xFFFFFFFF)
	{
		flog(LOG_ERR, "es1370: impossibile creare mutex");
		return false;
	}
	
	p_des->ctrl =	0x00000000;	// registro di controllo
	p_des->sctrl =	0x00000000;	// registro -seriale- di controllo
	p_des->cssr =	0x00000000;	// registro di stato
	p_des->uartc =	0x00;	// registro di controllo UART

	natb irq = pci_read(w, 0x3C, 1);
	p_des->irq=irq;	
	flog(LOG_INFO, "Piedino di interruzione numero %d", irq);

	if(activate_pe(estern_snd, 0, PRIO, LIV, irq)==0xFFFFFFFF)
	{
  		flog(LOG_ERR, "es1370: impossibile creare processo esterno");
		return false;	
	}
			
#if 0			
	if(!ak4531_init(base_address))
		return false;		
#endif

	for(int i=0; i<3; i++)
	{	if (((p_des->canali[i].mutex = sem_ini(1)) == 0xFFFFFFFF) || 
		    ((p_des->canali[i].sincr = sem_ini(0)) == 0xFFFFFFFF)){
		    		flog(LOG_ERR, "es1370: impossibile creare mutex");
				return false;
		    }
	}
	
	outputl(p_des->ctrl, p_des->indreg.control);
	outputl(p_des->sctrl, p_des->indreg.serial);
	outputl(p_des->cssr, p_des->indreg.status);
	outputb(p_des->uartc, p_des->indreg.uart_status);

	p_des->presente= true;
	return true;

}


// cerca se è installata una scheda audio e chiama la corrispondente 
// funzione di inizializzazione
// per ora si limita a controllare se la ES1370 è installata

bool find_sound()
{
	bool found = false;
	bool ciclo2;
	natb dev, fun;
	dev=0;
	fun=0;
	flog(LOG_INFO, "Ricerca scheda audio");
	natw vendorID, deviceID, w;
	for(; dev<32 && !found; dev++){
		ciclo2=true;	
		for(fun=0; fun <8 && ciclo2; fun++){
			
			w= make_w(dev, fun);
			vendorID= pci_read(w, 0, 2);

			if(vendorID == 0xFFFF) 
				ciclo2=false;

			if(vendorID == ES_VENDOR_ID){
				deviceID= pci_read(w, 2, 2);
				if(deviceID == ES_DEVICE_ID){	
					found = true;
					flog(LOG_INFO, "Rilevata scheda audio ES1370 sulla periferica %d, funzione %d", dev, fun);
					if(!ensoniq_init(dev, fun))
						flog(LOG_ERR, "Inizializzazione scheda audio fallita");
					else	
						flog(LOG_INFO, "Scheda audio ES1370 inizializzata correttamente");
				}
			}
		}
	}
	return found;
}

void rilascia_dac(int);


// gestione delle interruzioni per i canali di esecuzione
void gestisci_int_dac(int channel)
{	
	des_snd * p_des;
	p_des= &ensoniq;
	
	natl BUFFERLONG = BUFFERSIZE>>2;
			// numero delle parole lunghe contenute in un buffer
	natl appoggio;
	natw current;
	int difflong;
	int copia_n;
	ioaddr dac1_size= (channel==1)? p_des->indreg.dac1_size : 
								  p_des->indreg.dac2_size;
								  
	int oldsize=p_des->canali[channel].oldsize;
			// prendiamo l'indice dell'ultima copia nel buffer
	addr puntatore_brano=p_des->canali[channel].puntatore_brano;
	addr buffer=p_des->canali[channel].buffer;
	int ancora = p_des->canali[channel].quanti;
	

	sem_wait(p_des->es_mutex);
	outputl(0x0000000C, p_des->indreg.mem_page);
	inputl(dac1_size, appoggio);
	sem_signal(p_des->es_mutex);
	current = (appoggio& 0xFFFF0000)>>16;
			// in current ora c'è l'indice del buffer che è stato
			// copiato nella scheda audio
	
	difflong = (current > oldsize)? (current - oldsize) 
			: (BUFFERLONG) - oldsize + current;
			// il buffer è circolare, current potrebbe essere minore di
			// oldsize
	
	if(ancora<=0)
	{
		// abbiamo un numero di byte da scrivere negativo, ovvero
		// il brano è stato tutto copiato, ma non eseguito: 
		// riempiamo gradualmente il buffer di 0x00, ovvero silenzio
		// il loop farà il resto
		
		mute_snd_buffer(buffer, oldsize, difflong);
		p_des->canali[channel].quanti -= (difflong<<2);
		if(p_des->canali[channel].quanti<=(~BUFFERSIZE)+1)
			rilascia_dac(1);
		// se abbiamo quanti < -BUFFERSIZE (complemento a due)
		// significa che il buffer è stato riempito tutto di 0: 
		// possiamo staccare 
		
		p_des->canali[channel].oldsize = current;
		// altrimenti ricopiamo il valore di current in oldsize
		return;
	}
	
	copia_n= (difflong < (ancora>>2))? difflong:(ancora>>2);	
	copia_snd_buffer(puntatore_brano, buffer, oldsize, copia_n);
		// copiamo un numero di parole lunghe pari:
		// 	alle parole lunghe copiate dalla scheda audio o
		//	alle parole lunghe che rimangono nel brano
	
	p_des->canali[channel].puntatore_brano = (int*)puntatore_brano
												+ copia_n;		
	p_des->canali[channel].quanti -= (copia_n<<2);
	ancora = p_des->canali[channel].quanti;
		// aggiorniamo il puntatore all'interno del brano
		// e il contatore di byte del brano
		
	if(difflong>copia_n)
	{	
		// il brano era finito: abbiamo copiato le ultime long
		// dobbiamo ora copiare gli ultimi byte del brano
		// e riempire il resto del buffer di silenzio
		oldsize=(oldsize + copia_n)%BUFFERLONG;
		
		copia_snd_buffer_B(puntatore_brano, buffer, oldsize, ancora);
		copia_n = (difflong<<2) - (copia_n<<2) - ancora;
		oldsize= ((oldsize<<2) + copia_n)%BUFFERSIZE;
		mute_snd_buffer_B(buffer, oldsize, copia_n);
		p_des->canali[channel].quanti-=(ancora + copia_n);
	}	
	p_des->canali[channel].oldsize = current;
		// aggiorniamo l'indice all'interno del buffer
}	

	
// Processo esterno per la scheda audio
void estern_snd(int h)
{
	des_snd * p_des;
	p_des = & ensoniq;
	natl status;

	for(;;)
	{
		sem_wait(p_des->es_mutex);
		inputl(p_des->indreg.status, status);	
		sem_signal(p_des->es_mutex);

		if(!(status & ES1370_INTR))
		{	
			continue;
		}
		
		if (status & ES1370_ADC)
		;
		if (status & ES1370_DAC1)
		{
			sem_wait(p_des->es_mutex);
			p_des->sctrl &= ~(ES1370_P1_INT_EN);	
			outputl(p_des->sctrl, p_des->indreg.serial);
			p_des->sctrl |= (ES1370_P1_INT_EN);	
			outputl(p_des->sctrl, p_des->indreg.serial);
			sem_signal(p_des->es_mutex);
			gestisci_int_dac(1);
			// disabilitazione e riabilitazione delle interruzioni
			// e gestione della interruzione
		}	
		if (status & ES1370_DAC2)
		;
		wfi();
	}
}


// Preparazione del canale di esecuzione 1
void prepara_dac1(des_snd * p_des, addr puntatore, natl size, 
												natl rate, natl mode)
{
	sem_wait(p_des->es_mutex);
	p_des->ctrl &= ~(ES1370_DAC1_EN);
	outputl(p_des->ctrl, p_des->indreg.control);
	sem_signal(p_des->es_mutex);

	// disabilitamo il canale
	
	addr buffersuono = mem_alloc(BUFFERSIZE, BUFFERALIGN);
	// allochiamo il buffer
	
	if((buffersuono==0)||!resident(buffersuono, BUFFERSIZE, true))
	{
		flog(LOG_ERR, "Impossibile creare buffer sonoro");
		mem_free(buffersuono);
		terminate_p();
	}
	
	natb shift;
	
	switch(mode & 0x03){
		case 3: shift=2; break;
		case 2:
		case 1: shift=1; break;
		case 0: shift=0; break;
	}
	// shift rappresenta la grandezza di un frame in log2
	// mode = 00 -> modalità 8 bit mono: frame = 1 byte
	// mode = 01 -> modalità 8 bit stereo: frame = 2 byte
	// mode = 10 -> modalità 16 bit mono: frame = 2 byte
	// mode = 11 -> modalità 16 bit stereo: frame = 4 byte

	natl dac1_count = (BUFFERSIZE >> (shift+1));
	// diciamo alla scheda audio di lanciare un'interruzione dopo
	// aver eseguito mezzo buffer
	
	natb ratectrl;
	switch (rate) 
	{
		case 5512: ratectrl=0; break;
		case 11025: ratectrl=1; break;
		case 22050: ratectrl=2; break;
		case 44100: ratectrl=3; break;
	}
	
	natl dac1_size= BUFFERSIZE >> 2;
	
	addr ind_fisico= trasforma(buffersuono);
	memcpy(buffersuono, puntatore, BUFFERSIZE);
	
	p_des->canali[1].buffer = buffersuono;
	p_des->canali[1].puntatore_brano = (natb*)puntatore + BUFFERSIZE;
	p_des->canali[1].quanti = size - BUFFERSIZE;
	p_des->canali[1].oldsize=0;
	// salvataggio nel descrittore di canale
	// dell'indirizzo del buffer, il puntatore all'interno del brano
	// quantità dei dati da scrivere
	// e indice all'interno del buffer
	
	sem_wait(p_des->es_mutex);
	p_des->ctrl |= (ratectrl&0x03)<<12;
	outputl(p_des->ctrl, p_des->indreg.control);	
	outputl(0x0000000C, p_des->indreg.mem_page);	
	outputl(dac1_count-1, p_des->indreg.dac1_count);
	outputl(dac1_size-1, p_des->indreg.dac1_size);
	outputl(natl(ind_fisico), p_des->indreg.dac1_frame);
	
	p_des->sctrl &= ~(ES1370_P1_LOOP_SEL  | ES1370_P1_SCT_RLD | 0x03 | ES1370_P1_PAUSE);
			// Loop mode; continua a suonare, quando il count dei samples va a zero
				// lancia un'interruzione, se abilitato
			// da tenere basso per specifiche
			// 0x03 serve da maschera per la modalità del canale
			// metto a 0 il bit di pause, che equivale a mettere il canale in modalità
				// play
			
	p_des->sctrl |= (mode & 0x00000003) | ES1370_P1_INT_EN;
			// mode: modalità del canale
			// abilitazione a lanciare interruzioni			
	outputl(p_des->sctrl, p_des->indreg.serial);				
	sem_signal(p_des->es_mutex);
	
	// scrittura nella scheda audio dei parametri
}

// questa funzione serve a eseguire 3 comandi: play, stop e pause
// per ora implementa solo il canale di esecuzione 1
void es1370_ioctl(des_snd * p_des, int channel, int cmd)
{
	sem_wait(p_des->es_mutex);
	switch(channel)
	{
		case 0: break;
		
		case 1:
			p_des->ctrl &= ~(ES1370_DAC1_EN);
			outputl(p_des->ctrl, p_des->indreg.control);
			
			if(cmd != ES1370_STOP)
			{
				natl what=0;
				what |= ES1370_P1_PAUSE;
				if(cmd == ES1370_PAUSE)
				{	p_des->sctrl |= what;
				}
				else
				{
					p_des->sctrl &= ~what;					
				}	
				outputl(p_des->sctrl, p_des->indreg.serial);					
				p_des->ctrl |= (ES1370_DAC1_EN);
			}
			else
				p_des->ctrl &= ~(ES1370_P1_INT_EN);	
							
				outputl(p_des->ctrl, p_des->indreg.control);
			break;
		case 2: break;		
	}
	sem_signal(p_des->es_mutex);
}

// funzione chiamata alla fine della gestisci_int_dac
// quando cioè è stato scritto un buffer di silenzio
void rilascia_dac(int channel)
{
	des_snd * p_des;
	p_des= &ensoniq;	
	
	es1370_ioctl(p_des, channel, ES1370_STOP);	
		
	resident(p_des->canali[channel].buffer, BUFFERSIZE, false);
	mem_free(p_des->canali[channel].buffer);	
		//ripulitura buffer
	
	sem_signal(p_des->canali[channel].mutex);
	sem_signal(p_des->canali[channel].sincr);
	flog(LOG_INFO, "Canale %d rilasciato", channel);	
}


// primitive per il modulo utente
extern "C" void es1370_play(addr puntatore, natl size, natl rate, 
												natl mode, int channel)
{
	des_snd * p_des;
	p_des= &ensoniq;
	
	if(!p_des->presente)
	{
		flog(LOG_WARN, "Errore nell'acquisizione del canale: scheda audio non presente!");
		abort_p();
	}
	if(channel<=0 || channel>2)
	{
		flog(LOG_WARN, "Errore nell'acquisizione del canale: canale %d inesistente", channel);
		abort_p();
	}
	
	if(channel==1)
	{
		switch(rate)
		{
			case 5512: case 11025:
			case 22050: case 44100: break;
			default:
				flog(LOG_WARN, "Tentativo di eseguire un file con rate non valido per il canale DAC1");
				abort_p();
		}
			// Il canale DAC1 ha un rate fisso: se cerchiamo di eseguire un suono
			// con un rate diverso su quel canale, meglio se non iniziamo neanche a inizializzarlo
	}
	sem_wait(p_des->canali[channel].mutex);
			// semaforo di mutua esclusione sul canale: aspettiamo che
			// il canale venga liberato
	switch(channel)
	{
//		case 0: prepara adc(p_des, rate, mode); 
//				record adc(p_des, ind_fisico, size); break;
		case 1: prepara_dac1(p_des, puntatore, size, rate, mode); 
				es1370_ioctl(p_des, channel, ES1370_RESUME);
				
//		case 2: prepara_dac2(p_des, rate, mode); break;		
	}	
	flog(LOG_INFO, "Acquisito canale %d", channel);
	sem_wait(p_des->canali[channel].sincr);
			// aspettiamo che il brano sia finito (o che venga mandato
			// uno stop anzitempo)
}

extern "C" void es1370_command(int channel, int cmd)
{
	des_snd * p_des;
	p_des= &ensoniq;
	if(!p_des->presente)
	{
		flog(LOG_WARN, "Errore nell'invio dei comandi: scheda audio non presente!");
		abort_p();
	}
	
	if(cmd<0 || cmd > ES1370_PAUSE)
	{	flog(LOG_WARN, "Errore nella sound command");
		abort_p();
	}
	if(cmd!=ES1370_STOP)
		es1370_ioctl(p_des, channel, cmd);
	else
		if(channel!=0)
			rilascia_dac(channel);
			
	// la primitiva può fidarsi dei canali, 
	// tanto in caso siano sbagliati la ioctl non fa nulla
}

////////////////////////////////////////////////////////////////////////////////
//                 INIZIALIZZAZIONE DEL SOTTOSISTEMA DI I/O                   //
////////////////////////////////////////////////////////////////////////////////

// inizializza i gate usati per le chiamate di IO
//
extern "C" void fill_io_gates(void);

// eseguita in fase di inizializzazione
//
extern "C" void cmain(int sem_io)
{
	fill_io_gates();
	if (!kbd_init()) {
		flog(LOG_WARN, "kdb_init fallita");
		abort_p();
	}
	if (!vkbd_init()) {
		flog(LOG_WARN, "vkdb_init fallita");
		abort_p();
	}
	if (!com_init()) {
		flog(LOG_WARN, "com_init fallita");
		abort_p();
	}
	if (!vmon_init()) {
		flog(LOG_WARN, "vmon_init fallita");
		abort_p();
	}
	if (!console_init(VKBD_CONSOLE, 0)) {
		flog(LOG_WARN, "console_init fallita");
		abort_p();
	}

	if (!find_sound()) {
		flog(LOG_WARN, "Scheda audio non trovata");
		// non è necessario abortire il processo:
		// il sistema operativo vive benissimo anche senza
		// scheda audio
	}

	sem_signal(sem_io);
	terminate_p();
}
